@@ -5,11 +5,13 @@ from beartype.typing import Optional, Union, List
 from beartype import beartype
 
 import torch
-from torch import nn, einsum
+from torch import nn, einsum, Tensor
 from torch.autograd import grad as torch_grad
 import torch.nn.functional as F
 from torch.nn.utils.rnn import pad_sequence
 
+import torchaudio
+
 from einops import rearrange, repeat, reduce
 from einops.layers.torch import Rearrange
 
@@ -23,8 +25,12 @@ from torchaudio.functional import resample
 from audiolm_pytorch.soundstream import SoundStream
 from audiolm_pytorch.encodec import EncodecWrapper
 from audiolm_pytorch.utils import AudioConditionerBase
+from audiolm_pytorch.attend import Attend
 
 from tqdm import tqdm
+from pathlib import Path
+from audiolm_pytorch.version import __version__
+from packaging import version
 
 # helper functions
 
@@ -220,17 +226,6 @@ class RelativePositionBias(nn.Module):
 
 # feedforward
 
-class CausalDSConv(nn.Module):
-    def __init__(self, dim):
-        super().__init__()
-        self.ds_conv = nn.Conv1d(dim, dim, 3, bias = False, groups = dim)
-
-    def forward(self, x):
-        x = rearrange(x, 'b n c -> b c n')
-        x = F.pad(x, (2, 0))
-        x = self.ds_conv(x)
-        return rearrange(x, 'b c n -> b n c')
-
 class GEGLU(nn.Module):
     def forward(self, x):
         x, gate = x.chunk(2, dim = -1)
@@ -241,7 +236,6 @@ def FeedForward(dim, mult = 4, dropout = 0.1):
     return nn.Sequential(
         LayerNorm(dim),
         nn.Linear(dim, inner_dim * 2, bias = False),
-        CausalDSConv(inner_dim * 2),
         GEGLU(),
         LayerNorm(inner_dim),
         nn.Dropout(dropout),
@@ -261,11 +255,11 @@ class Attention(nn.Module):
         norm_context = False,
         num_null_kv = 0,
         dropout = 0.1,
-        scale = 8
+        scale = 8,
+        flash = False
     ):
         super().__init__()
         self.heads = heads
-        self.scale = scale
         self.causal = causal
         inner_dim = dim_head * heads
 
@@ -277,13 +271,16 @@ class Attention(nn.Module):
         self.attn_dropout = nn.Dropout(dropout)
 
         self.num_null_kv = num_null_kv
-        self.null_kv = nn.Parameter(torch.randn(2, num_null_kv, dim_head))
+        self.null_kv = nn.Parameter(torch.randn(2, num_null_kv, dim_head)) if num_null_kv > 0 else None
 
         self.to_q = nn.Linear(dim, inner_dim, bias = False)
         self.to_kv = nn.Linear(dim_context, dim_head * 2, bias = False)
 
-        self.q_scale = nn.Parameter(torch.ones(dim_head))
-        self.k_scale = nn.Parameter(torch.ones(dim_head))
+        self.attend = Attend(
+            flash = flash,
+            dropout = dropout,
+            causal = causal
+        )
 
         self.to_out = nn.Sequential(
             nn.Linear(inner_dim, dim, bias = False),
@@ -343,38 +340,14 @@ class Attention(nn.Module):
 
         q = rearrange(q, 'b n (h d) -> b h n d', h = self.heads)
 
-        # new technique, rmsnormed queries and keys, first used by 22B parameter model successfully https://arxiv.org/abs/2302.05442
-
-        q, k = map(l2norm, (q, k))
-        q = q * self.q_scale
-        k = k * self.k_scale
-
-        # similarities
-
-        sim = einsum('b h i d, b j d -> b h i j', q, k) * self.scale
-
-        if exists(attn_bias):
-            attn_bias = F.pad(attn_bias, (self.num_null_kv, 0), value = 0.)
-            sim = sim + attn_bias
+        # handle mask and null key / value
 
         if exists(mask):
             mask = F.pad(mask, (self.num_null_kv, 0), value = True)
-            mask = rearrange(mask, 'b j -> b 1 1 j')
-            sim = sim.masked_fill(~mask, -torch.finfo(sim.dtype).max)
-
-        if self.causal:
-            i, j = sim.shape[-2:]
-            causal_mask = torch.ones((i, j), dtype = torch.bool, device = x.device).triu(j - i + 1)
-            sim = sim.masked_fill(causal_mask, -torch.finfo(sim.dtype).max)
 
         # attention
 
-        attn = sim.softmax(dim = -1)
-        attn = self.attn_dropout(attn)
-
-        # aggregate
-
-        out = einsum('b h i j, b j d -> b h i d', attn, v)
+        out = self.attend(q, k, v, attn_bias = attn_bias, mask = mask)
 
         # merge heads
 
@@ -397,10 +370,14 @@ class Transformer(nn.Module):
         grad_shrink_alpha = 0.1,
         cond_as_self_attn_prefix = False,
         rel_pos_bias = True,
+        flash_attn = False,
         **kwargs
     ):
         super().__init__()
+        rel_pos_bias = rel_pos_bias and not flash_attn
+
         assert not (cross_attend and cond_as_self_attn_prefix)
+
         self.dim_context = default(dim_context, dim)
 
         self.cond_as_self_attn_prefix = cond_as_self_attn_prefix
@@ -413,8 +390,8 @@ class Transformer(nn.Module):
 
         for _ in range(depth):
             self.layers.append(nn.ModuleList([
-                Attention(dim = dim, heads = heads, dropout = attn_dropout, causal = True, **kwargs),
-                Attention(dim = dim, heads = heads, dropout = attn_dropout, dim_context = dim_context, num_null_kv = 1, norm_context = True, **kwargs) if cross_attend else None,
+                Attention(dim = dim, heads = heads, dropout = attn_dropout, flash = flash_attn, causal = True, **kwargs),
+                Attention(dim = dim, heads = heads, dropout = attn_dropout, dim_context = dim_context, flash = flash_attn, num_null_kv = 1, norm_context = True, **kwargs) if cross_attend else None,
                 FeedForward(dim = dim, dropout = ff_dropout)
             ]))
 
@@ -479,9 +456,13 @@ class SemanticTransformer(nn.Module):
         cond_as_self_attn_prefix = False,
         cond_drop_prob = 0.5,
         grad_shrink_alpha = 0.1,
+        rel_pos_bias = True,
+        flash_attn = False,
         **kwargs
     ):
         super().__init__()
+        rel_pos_bias = rel_pos_bias and not flash_attn
+
         self.num_semantic_tokens = num_semantic_tokens
 
         if audio_text_condition:
@@ -509,6 +490,8 @@ class SemanticTransformer(nn.Module):
             cross_attend = has_condition and not cond_as_self_attn_prefix,
             cond_as_self_attn_prefix = cond_as_self_attn_prefix,
             grad_shrink_alpha = grad_shrink_alpha,
+            rel_pos_bias = rel_pos_bias,
+            flash_attn = flash_attn,
             **kwargs
         )
 
@@ -518,6 +501,19 @@ class SemanticTransformer(nn.Module):
     def device(self):
         return next(self.parameters()).device
 
+    def load(self, path):
+        # Return pkg so that if this function gets called from within a Trainer function call,
+        # the trainer can also access the package loaded from the checkpoint.
+        device = self.device
+        path = Path(path)
+        assert path.exists()
+        pkg = torch.load(str(path), map_location = device)
+        # check version
+        if 'version' in pkg and version.parse(pkg['version']) < version.parse(__version__):
+            print(f'model was trained on older version {pkg["version"]} of audiolm-pytorch')
+        self.load_state_dict(pkg['model'])
+        return pkg
+
     def forward_with_cond_scale(
         self,
         *args,
@@ -553,7 +549,7 @@ class SemanticTransformer(nn.Module):
 
         text_mask = None
         if not exists(text_embeds) and exists(text):
-            with torch.no_grad():
+            with torch.inference_mode():
                 text_embeds = self.embed_text(text, output_device = device)
                 text_mask = torch.any(text_embeds != 0, dim = -1)
 
@@ -602,9 +598,13 @@ class CoarseTransformer(nn.Module):
         cond_drop_prob = 0.5,
         grad_shrink_alpha = 0.1,
         project_semantic_logits = True,
+        rel_pos_bias = True,
+        flash_attn = False,
         **kwargs
     ):
         super().__init__()
+        rel_pos_bias = rel_pos_bias and not flash_attn
+
         self.num_semantic_tokens = num_semantic_tokens
 
         if audio_text_condition:
@@ -630,7 +630,7 @@ class CoarseTransformer(nn.Module):
         text_dim = default(cond_dim, get_encoded_dim(t5_name))
         self.proj_text_embed = nn.Linear(text_dim, dim, bias = False) if text_dim != dim else nn.Identity()
 
-        self.cross_attn_bias = nn.Parameter(torch.zeros(heads, 1, 1))
+        self.cross_attn_bias = nn.Parameter(torch.zeros(heads, 1, 1)) if rel_pos_bias else None
 
         self.transformer = Transformer(
             dim = dim,
@@ -641,6 +641,8 @@ class CoarseTransformer(nn.Module):
             cross_attend = has_condition and not cond_as_self_attn_prefix,
             cond_as_self_attn_prefix = cond_as_self_attn_prefix,
             grad_shrink_alpha = grad_shrink_alpha,
+            rel_pos_bias = rel_pos_bias,
+            flash_attn = flash_attn,
             **kwargs
         )
 
@@ -654,6 +656,19 @@ class CoarseTransformer(nn.Module):
     def device(self):
         return next(self.parameters()).device
 
+    def load(self, path):
+        # Return pkg so that if this function gets called from within a Trainer function call,
+        # the trainer can also access the package loaded from the checkpoint.
+        device = self.device
+        path = Path(path)
+        assert path.exists()
+        pkg = torch.load(str(path), map_location = device)
+        # check version
+        if 'version' in pkg and version.parse(pkg['version']) < version.parse(__version__):
+            print(f'model was trained on older version {pkg["version"]} of audiolm-pytorch')
+        self.load_state_dict(pkg['model'])
+        return pkg
+
     def forward_with_cond_scale(
         self,
         *args,
@@ -693,7 +708,7 @@ class CoarseTransformer(nn.Module):
         assert not (self.has_condition ^ has_text)
 
         if not exists(text_embeds) and exists(text):
-            with torch.no_grad():
+            with torch.inference_mode():
                 text_embeds = self.embed_text(text, output_device = device)
 
         text_mask = None
@@ -737,16 +752,20 @@ class CoarseTransformer(nn.Module):
         # engineer the attention bias so that cross attention is not dominated by relative positions
 
         seq_len = tokens.shape[-2]
-        attn_bias = self.transformer.rel_pos_bias(seq_len)
 
-        is_semantic = arange(seq_len) < (semantic_seq_len + 1) # semantic seq len + start token
-        is_cross_attn = rearrange(is_semantic, 'i -> i 1') ^ rearrange(is_semantic, 'j -> 1 j')
+        attn_bias = None
 
-        attn_bias = torch.where(
-            is_cross_attn,
-            self.cross_attn_bias,
-            attn_bias
-        )
+        if exists(self.transformer.rel_pos_bias):
+            attn_bias = self.transformer.rel_pos_bias(seq_len)
+
+            is_semantic = arange(seq_len) < (semantic_seq_len + 1) # semantic seq len + start token
+            is_cross_attn = rearrange(is_semantic, 'i -> i 1') ^ rearrange(is_semantic, 'j -> 1 j')
+
+            attn_bias = torch.where(
+                is_cross_attn,
+                self.cross_attn_bias,
+                attn_bias
+            )
 
         # attend
 
@@ -808,9 +827,13 @@ class FineTransformer(nn.Module):
         cond_drop_prob = 0.5,
         grad_shrink_alpha = 0.1,
         project_coarse_logits = True,
+        pad_id = -1,
+        rel_pos_bias = True,
+        flash_attn = False,
         **kwargs
     ):
         super().__init__()
+        rel_pos_bias = rel_pos_bias and not flash_attn
 
         if audio_text_condition:
             has_condition = True
@@ -831,6 +854,7 @@ class FineTransformer(nn.Module):
         self.coarse_quantize_embedding = nn.Embedding(num_coarse_quantizers, dim)
         self.fine_quantize_embedding = nn.Embedding(num_fine_quantizers, dim)
 
+        self.pad_id = pad_id
         self.eos_id = codebook_size
 
         text_dim = default(cond_dim, get_encoded_dim(t5_name))
@@ -846,21 +870,23 @@ class FineTransformer(nn.Module):
             cond_as_self_attn_prefix = cond_as_self_attn_prefix,
             rel_pos_bias = False,
             grad_shrink_alpha = grad_shrink_alpha,
+            flash_attn = flash_attn,
             **kwargs
         )
 
         # doing a specialized attn bias so that corresponding time steps at fine and coarse sequences attend to each other better
 
-        self.null_pos_bias = nn.Parameter(torch.randn(heads, 1, 1))
+        self.null_pos_bias = nn.Parameter(torch.randn(heads, 1, 1)) if rel_pos_bias else None
 
         pos_bias_mlp_dim = dim // 2
+
         self.pos_bias_mlp = nn.Sequential(
             nn.Linear(2, pos_bias_mlp_dim),
             nn.SiLU(),
             nn.Linear(pos_bias_mlp_dim, pos_bias_mlp_dim),
             nn.SiLU(),
             nn.Linear(pos_bias_mlp_dim, heads)
-        )
+        ) if rel_pos_bias else None
 
         self.codebook_size = codebook_size
         self.num_coarse_quantizers = num_coarse_quantizers
@@ -873,6 +899,20 @@ class FineTransformer(nn.Module):
     def device(self):
         return next(self.parameters()).device
 
+    def load(self, path):
+        # Return pkg so that if this function gets called from within a Trainer function call,
+        # the trainer can also access the package loaded from the checkpoint.
+        device = self.device
+        path = Path(path)
+        assert path.exists()
+        pkg = torch.load(str(path), map_location = device)
+        # check version
+        if 'version' in pkg and version.parse(pkg['version']) < version.parse(__version__):
+            print(f'model was trained on older version {pkg["version"]} of audiolm-pytorch')
+        self.load_state_dict(pkg['model'])
+        return pkg
+
+
     def forward_with_cond_scale(
         self,
         *args,
@@ -904,12 +944,15 @@ class FineTransformer(nn.Module):
         return_only_fine_logits = False
     ):
         b, device = coarse_token_ids.shape[0], coarse_token_ids.device
+
+        # handle text conditioning
+
         has_text = exists(text) or exists(text_embeds)
         assert not (self.has_condition ^ has_text)
 
         text_mask = None
         if not exists(text_embeds) and exists(text):
-            with torch.no_grad():
+            with torch.inference_mode():
                 text_embeds = self.embed_text(text, output_device = device)
                 text_mask = torch.any(text_embeds != 0, dim = -1)
 
@@ -924,6 +967,21 @@ class FineTransformer(nn.Module):
 
         coarse_token_ids, fine_token_ids = map(lambda t: rearrange(t, 'b ... -> b (...)'), (coarse_token_ids, fine_token_ids))
 
+        # do not attend to any of the coarse padding tokens or coarse end token either
+
+        coarse_self_attn_mask = (coarse_token_ids != self.pad_id) & (coarse_token_ids != self.eos_id)
+        coarse_token_ids = coarse_token_ids.masked_fill(~coarse_self_attn_mask, 0)
+
+        fine_token_seq_len = fine_token_ids.shape[-1]
+        coarse_self_attn_mask = F.pad(coarse_self_attn_mask, (1, fine_token_seq_len + 1), value = True)
+
+        if exists(self_attn_mask):
+            self_attn_mask &= coarse_self_attn_mask
+        else:
+            self_attn_mask = coarse_self_attn_mask
+
+        # prepare coarse and fine token embeddings
+
         b, n = coarse_token_ids.shape
 
         coarse_length = coarse_token_ids.shape[-1]
@@ -963,75 +1021,78 @@ class FineTransformer(nn.Module):
 
         # an engineered attention bias so coarse and fine sequences attend to each other better
 
-        max_seq_len = max(coarse_seq_length, fine_seq_length)
+        attn_bias = None
 
-        coarse_pos = torch.arange(coarse_seq_length, device = device)
-        fine_pos = torch.arange(fine_seq_length, device = device)
+        if exists(self.pos_bias_mlp):
+            max_seq_len = max(coarse_seq_length, fine_seq_length)
 
-        coarse_pos = repeat(coarse_pos, 'n -> (n q)', q = self.num_coarse_quantizers)[:coarse_length]
-        fine_pos = repeat(fine_pos, 'n -> (n q)', q = self.num_fine_quantizers)[:fine_length]
+            coarse_pos = torch.arange(coarse_seq_length, device = device)
+            fine_pos = torch.arange(fine_seq_length, device = device)
 
-        coarse_pos = F.pad(coarse_pos, (1, 0), value = -1)
-        fine_pos = F.pad(fine_pos, (1, 0), value = -1)
+            coarse_pos = repeat(coarse_pos, 'n -> (n q)', q = self.num_coarse_quantizers)[:coarse_length]
+            fine_pos = repeat(fine_pos, 'n -> (n q)', q = self.num_fine_quantizers)[:fine_length]
 
-        seq_positions = torch.cat((coarse_pos, fine_pos), dim = -1)
+            coarse_pos = F.pad(coarse_pos, (1, 0), value = -1)
+            fine_pos = F.pad(fine_pos, (1, 0), value = -1)
 
-        coarse_offsets = F.pad(coarse_offsets, (1, 0), value = 0)
-        fine_offsets = fine_offsets + self.num_coarse_quantizers
-        fine_offsets = F.pad(fine_offsets, (1, 0), value = 0)
+            seq_positions = torch.cat((coarse_pos, fine_pos), dim = -1)
 
-        seq_offsets = torch.cat((coarse_offsets, fine_offsets), dim = -1)
+            coarse_offsets = F.pad(coarse_offsets, (1, 0), value = 0)
+            fine_offsets = fine_offsets + self.num_coarse_quantizers
+            fine_offsets = F.pad(fine_offsets, (1, 0), value = 0)
 
-        pos_mlp_input = torch.stack((seq_positions.clamp(min = 0), seq_offsets), dim = -1)
+            seq_offsets = torch.cat((coarse_offsets, fine_offsets), dim = -1)
 
-        num_offsets = self.num_fine_quantizers + self.num_coarse_quantizers
+            pos_mlp_input = torch.stack((seq_positions.clamp(min = 0), seq_offsets), dim = -1)
 
-        # relative positions are always (2 * N - 1), where N is the length of the dimension
+            num_offsets = self.num_fine_quantizers + self.num_coarse_quantizers
 
-        rel_seq_len, rel_offsets = map(lambda n: 2 * n - 1, (max_seq_len, num_offsets))
+            # relative positions are always (2 * N - 1), where N is the length of the dimension
 
-        # get all relative distances
+            rel_seq_len, rel_offsets = map(lambda n: 2 * n - 1, (max_seq_len, num_offsets))
 
-        rel_dist = (rearrange(pos_mlp_input, 'i c -> i 1 c') - rearrange(pos_mlp_input, 'j c -> 1 j c'))
+            # get all relative distances
 
-        # get all possible relative distances for the attention bias to be computed from the mlp
-        # which would be - (2 * N - 1) * (2 * Q - 1) - where N = sequence length and Q = total quantizers
+            rel_dist = (rearrange(pos_mlp_input, 'i c -> i 1 c') - rearrange(pos_mlp_input, 'j c -> 1 j c'))
 
-        rel_seq_len_range = repeat(torch.arange(rel_seq_len, device = device), 'n -> (n q)', q = rel_offsets)
-        rel_offset_range = repeat(torch.arange(rel_offsets, device = device), 'q -> (n q)', n = rel_seq_len)
+            # get all possible relative distances for the attention bias to be computed from the mlp
+            # which would be - (2 * N - 1) * (2 * Q - 1) - where N = sequence length and Q = total quantizers
 
-        mlp_inputs = torch.stack((rel_seq_len_range, rel_offset_range), dim = -1)
+            rel_seq_len_range = repeat(torch.arange(rel_seq_len, device = device), 'n -> (n q)', q = rel_offsets)
+            rel_offset_range = repeat(torch.arange(rel_offsets, device = device), 'q -> (n q)', n = rel_seq_len)
 
-        # implicitly parameterized relative distances, by sequence and quantizer positions
+            mlp_inputs = torch.stack((rel_seq_len_range, rel_offset_range), dim = -1)
 
-        attn_bias = self.pos_bias_mlp(mlp_inputs.float())
+            # implicitly parameterized relative distances, by sequence and quantizer positions
 
-        # translate coordinates of (rel_seq_pos, rel_quantizer_offset) -> positive index to select from attn bias
+            attn_bias = self.pos_bias_mlp(mlp_inputs.float())
 
-        rel_dist_seq_pos, rel_dist_seq_offset = rel_dist.unbind(dim = -1)
+            # translate coordinates of (rel_seq_pos, rel_quantizer_offset) -> positive index to select from attn bias
 
-        rel_dist_seq_pos += max_seq_len - 1
-        rel_dist_seq_offset += num_offsets - 1
+            rel_dist_seq_pos, rel_dist_seq_offset = rel_dist.unbind(dim = -1)
 
-        rel_dist_indices = rel_dist_seq_pos * rel_offsets + rel_dist_seq_offset
+            rel_dist_seq_pos += max_seq_len - 1
+            rel_dist_seq_offset += num_offsets - 1
 
-        # select the relative positional attention bias outputted by the MLP
-        # savings go from (N * Q) ^ 2 -> ~ (4 * N * Q)
+            rel_dist_indices = rel_dist_seq_pos * rel_offsets + rel_dist_seq_offset
 
-        attn_bias = attn_bias[rel_dist_indices]
+            # select the relative positional attention bias outputted by the MLP
+            # savings go from (N * Q) ^ 2 -> ~ (4 * N * Q)
 
-        attn_bias = rearrange(attn_bias, '... h -> h ...')
+            attn_bias = attn_bias[rel_dist_indices]
 
-        # need to make sure start token has a custom positional bias
+            attn_bias = rearrange(attn_bias, '... h -> h ...')
 
-        is_start_token_seq = seq_positions == -1
-        start_token_mask = rearrange(is_start_token_seq, 'i -> i 1') | rearrange(is_start_token_seq, 'j -> 1 j')
+            # need to make sure start token has a custom positional bias
 
-        attn_bias = torch.where(
-            start_token_mask,
-            self.null_pos_bias,
-            attn_bias,
-        )
+            is_start_token_seq = seq_positions == -1
+            start_token_mask = rearrange(is_start_token_seq, 'i -> i 1') | rearrange(is_start_token_seq, 'j -> 1 j')
+
+            attn_bias = torch.where(
+                start_token_mask,
+                self.null_pos_bias,
+                attn_bias,
+            )
 
         # attention
 
@@ -1106,6 +1167,7 @@ class SemanticTransformerWrapper(nn.Module):
         super().__init__()
         self.wav2vec = wav2vec
         self.transformer = transformer
+        self.to(transformer.device)
         self.audio_conditioner = audio_conditioner
 
         assert not (exists(audio_conditioner) and not transformer.has_condition), 'if conditioning on audio embeddings from mulan, transformer has_condition must be set to True'
@@ -1121,8 +1183,11 @@ class SemanticTransformerWrapper(nn.Module):
     def device(self):
         return next(self.parameters()).device
 
+    def embed_text(self, text):
+        return self.transformer.embed_text(text, output_device = self.device)
+
     @eval_decorator
-    @torch.no_grad()
+    @torch.inference_mode()
     @beartype
     def generate(
         self,
@@ -1131,6 +1196,7 @@ class SemanticTransformerWrapper(nn.Module):
         text: Optional[List[str]] = None,
         text_embeds = None,
         prime_wave = None,
+        prime_wave_input_sample_hz = None,
         prime_ids = None,
         batch_size = 1,
         cond_scale = 3,
@@ -1146,7 +1212,11 @@ class SemanticTransformerWrapper(nn.Module):
         if exists(prime_wave):
             assert not exists(prime_ids)
             assert exists(self.wav2vec)
-            ids = self.wav2vec(prime_wave, flatten = False)
+            ids = self.wav2vec(
+                prime_wave,
+                flatten = False,
+                input_sample_hz = prime_wave_input_sample_hz
+            )
         elif exists(prime_ids):
             ids = prime_ids
         else:
@@ -1167,7 +1237,7 @@ class SemanticTransformerWrapper(nn.Module):
         assert not (self.transformer.has_condition ^ has_text)
 
         if not exists(text_embeds) and exists(text):
-            with torch.no_grad():
+            with torch.inference_mode():
                 text_embeds = self.transformer.embed_text(text, output_device = device)
 
         # start length and get running id output
@@ -1284,6 +1354,7 @@ class CoarseTransformerWrapper(nn.Module):
         self.wav2vec = wav2vec
 
         self.transformer = transformer
+        self.to(transformer.device)
         self.audio_conditioner = audio_conditioner
 
         assert not (exists(audio_conditioner) and not transformer.has_condition), 'if conditioning on audio embeddings from mulan, transformer has_condition must be set to True'
@@ -1304,12 +1375,15 @@ class CoarseTransformerWrapper(nn.Module):
         return next(self.parameters()).device
 
     @eval_decorator
-    @torch.no_grad()
+    @torch.inference_mode()
     @beartype
     def generate(
         self,
         *,
         semantic_token_ids,
+        prime_wave: Optional[Tensor] = None,
+        prime_wave_input_sample_hz = None,
+        prime_coarse_token_ids: Optional[Tensor] = None,
         text: Optional[List[str]] = None,
         text_embeds = None,
         max_time_steps = 512,
@@ -1323,7 +1397,28 @@ class CoarseTransformerWrapper(nn.Module):
 
         semantic_token_ids = semantic_token_ids.to(device)
 
-        coarse_token_ids = torch.empty((batch, 0), device = device, dtype = torch.long)
+        # initialize coarse token ids
+        # if a prime audio wave was supplied, then start off with appropriate acoustic tokens
+
+        assert not (exists(prime_wave) and exists(prime_coarse_token_ids)), 'you can either pass in the prime as a raw wave (codec required) or as preprocessed acoustic token ids'
+
+        if exists(prime_coarse_token_ids):
+            coarse_token_ids = prime_coarse_token_ids
+        elif exists(prime_wave):
+            assert exists(self.codec)
+            with torch.inference_mode():
+                self.codec.eval()
+
+                _, indices, _ = self.codec(
+                    prime_wave,
+                    return_encoded = True,
+                    input_sample_hz = prime_wave_input_sample_hz
+                )
+
+                coarse_token_ids = indices[..., :self.num_coarse_quantizers]
+                coarse_token_ids = rearrange(coarse_token_ids, 'b ... -> b (...)')
+        else:
+            coarse_token_ids = torch.empty((batch, 0), device = device, dtype = torch.long)
 
         # derive text embeddings if needed
 
@@ -1331,7 +1426,7 @@ class CoarseTransformerWrapper(nn.Module):
         assert not (self.transformer.has_condition ^ has_text)
 
         if not exists(text_embeds) and exists(text):
-            with torch.no_grad():
+            with torch.inference_mode():
                 text_embeds = self.transformer.embed_text(text, output_device = device)
 
         if self.unique_consecutive:
@@ -1339,12 +1434,12 @@ class CoarseTransformerWrapper(nn.Module):
 
         # initialize
 
-        init_coarse_time_step = coarse_token_ids.shape[-1]
+        init_coarse_time_step = 0
         sampled_coarse_token_ids = coarse_token_ids.clone()
 
         for time_step in tqdm(range(init_coarse_time_step, max_time_steps), desc = 'generating coarse'):
             for ind in range(self.num_coarse_quantizers):
-                is_last_step = ind == (self.num_coarse_quantizers - 1)
+                just_finished_quantizer_step = (ind == 0 and time_step > 0)
 
                 _, coarse_logits = self.transformer.forward_with_cond_scale(
                     coarse_token_ids = sampled_coarse_token_ids,
@@ -1357,8 +1452,8 @@ class CoarseTransformerWrapper(nn.Module):
 
                 last_coarse_logits = coarse_logits[:, -1]
 
-                if not is_last_step:
-                    last_coarse_logits[:, -1] = float('-inf') # prevent from eos if not last quantizer step, but move this to masking logic within the transformer at some point, for both training and eval
+                if not just_finished_quantizer_step:
+                    last_coarse_logits[:, -1] = float('-inf') # prevent from eos in the middle of a time step
 
                 filtered_logits = top_k(last_coarse_logits, thres = filter_thres)
                 sampled = gumbel_sample(filtered_logits, temperature = temperature, dim = -1)
@@ -1408,15 +1503,17 @@ class CoarseTransformerWrapper(nn.Module):
         if not exists(coarse_token_ids):
             assert exists(self.codec), 'Codec must be provided if given raw wave for training'
 
-            with torch.no_grad():
+            with torch.inference_mode():
                 self.codec.eval()
                 _, indices, _ = self.codec(raw_wave_for_codec, return_encoded = True)
-                batch = raw_wave_for_codec.shape[0]
-                num_timesteps = raw_wave_for_codec.shape[1]
+
+                batch, num_timesteps = raw_wave_for_codec.shape
                 num_frames = int(num_timesteps / self.codec.seq_len_multiple_of)
+
                 assert indices.shape[0] == batch and indices.shape[1] == num_frames, \
                     f'Expected indices to have shape (batch, num_frames, num_coarse_quantizers + num_fine_quantizers), but got {indices.shape}'
-                coarse_token_ids, _ = indices[..., :self.num_coarse_quantizers], indices[..., self.num_coarse_quantizers:]
+
+                coarse_token_ids = indices[..., :self.num_coarse_quantizers]
 
         semantic_token_ids = rearrange(semantic_token_ids, 'b ... -> b (...)')
         coarse_token_ids = rearrange(coarse_token_ids, 'b ... -> b (...)')
@@ -1506,6 +1603,7 @@ class FineTransformerWrapper(nn.Module):
         self.codec = codec
 
         self.transformer = transformer
+        self.to(transformer.device)
         self.audio_conditioner = audio_conditioner
 
         assert not (exists(audio_conditioner) and not transformer.has_condition), 'if conditioning on audio embeddings from mulan, transformer has_condition must be set to True'
@@ -1530,12 +1628,15 @@ class FineTransformerWrapper(nn.Module):
         return next(self.parameters()).device
 
     @eval_decorator
-    @torch.no_grad()
+    @torch.inference_mode()
     @beartype
     def generate(
         self,
         *,
         coarse_token_ids,
+        prime_wave: Optional[Tensor] = None,
+        prime_wave_input_sample_hz = None,
+        prime_fine_token_ids: Optional[Tensor] = None,
         text: Optional[List[str]] = None,
         text_embeds = None,
         cond_scale = 3.,
@@ -1557,21 +1658,41 @@ class FineTransformerWrapper(nn.Module):
         assert not (self.transformer.has_condition ^ has_text)
 
         if not exists(text_embeds) and exists(text):
-            with torch.no_grad():
+            with torch.inference_mode():
                 text_embeds = self.transformer.embed_text(text, output_device = device)
 
-        # initialize
+        # initialize fine token ids
+        # if a prime wave was supplied, start off with fine acoustic tokens
 
-        fine_token_ids = torch.empty((batch, 0), device = device, dtype = torch.long)
+        assert not (exists(prime_wave) and exists(prime_fine_token_ids)), 'you can either pass in the prime as a raw wave (codec required) or as preprocessed acoustic token ids'
 
-        init_fine_time_step = fine_token_ids.shape[-1]
+        if exists(prime_fine_token_ids):
+            fine_token_ids = prime_fine_token_ids
+        elif exists(prime_wave):
+            assert exists(self.codec)
+            with torch.inference_mode():
+                self.codec.eval()
+                _, token_ids, _ = self.codec(
+                    prime_wave,
+                    return_encoded = True,
+                    input_sample_hz = prime_wave_input_sample_hz
+                )
+
+            fine_token_ids = token_ids[..., self.num_coarse_quantizers:]
+            fine_token_ids = rearrange(fine_token_ids, 'b ... -> b (...)')
+        else:
+            fine_token_ids = torch.empty((batch, 0), device = device, dtype = torch.long)
+
+        # calculate number of sampling steps
+
+        init_fine_time_step = fine_token_ids.shape[-1] // self.num_fine_quantizers
         max_time_steps = coarse_token_ids.shape[1] // self.num_coarse_quantizers
 
         sampled_fine_token_ids = fine_token_ids.clone()
 
         for time_step in tqdm(range(init_fine_time_step, max_time_steps), desc = 'generating fine'):
             for ind in range(self.num_fine_quantizers):
-                is_last_step = ind == (self.num_fine_quantizers - 1)
+                just_finished_quantizer_step = (ind == 0 and time_step > 0)
 
                 _, fine_logits = self.transformer.forward_with_cond_scale(
                     coarse_token_ids = coarse_token_ids,
@@ -1584,6 +1705,9 @@ class FineTransformerWrapper(nn.Module):
 
                 last_fine_logits = fine_logits[:, -1]
 
+                if not just_finished_quantizer_step:
+                    last_fine_logits[:, -1] = float('-inf')  # prevent from eos in the middle of a time step
+
                 filtered_logits = top_k(last_fine_logits, thres = filter_thres)
                 sampled = gumbel_sample(filtered_logits, temperature = temperature, dim = -1)
 
@@ -1639,12 +1763,13 @@ class FineTransformerWrapper(nn.Module):
         if exists(raw_wave):
             assert exists(self.codec), 'Codec must be provided if given raw wave for training'
 
-            with torch.no_grad():
+            with torch.inference_mode():
                 self.codec.eval()
                 _, token_ids, _ = self.codec(raw_wave, return_encoded = True)
-                batch = raw_wave.shape[0]
-                num_timesteps = raw_wave.shape[1]
+
+                batch, num_timesteps = raw_wave.shape
                 num_frames = int(num_timesteps / self.codec.seq_len_multiple_of)
+
                 assert token_ids.shape == torch.Size((batch, num_frames, self.num_coarse_quantizers + self.num_fine_quantizers)), \
                     f'Expected token ids to have shape (batch, num_frames, num_coarse_quantizers + num_fine_quantizers), but got {token_ids.shape}'
 
@@ -1661,18 +1786,17 @@ class FineTransformerWrapper(nn.Module):
             fine_labels = fine_token_ids
             fine_token_ids = fine_token_ids[:, :-1]
 
-        # do not attend to any of the coarse padding tokens or coarse end token either
-
-        self_attn_mask = coarse_token_ids != self.pad_id
-        coarse_token_ids = coarse_token_ids.masked_fill(~self_attn_mask, 0)
-
-        fine_token_seq_len = fine_token_ids.shape[-1]
-        self_attn_mask = F.pad(self_attn_mask, (1, fine_token_seq_len + 1), value = True)
-
         # forgetful causal mask - structured dropout
 
+        self_attn_mask = None
+
         if self.mask_prob > 0 and self.training:
-            self_attn_mask &= generate_mask_with_prob(self_attn_mask.shape, self.mask_prob, device = self_attn_mask.device)
+            mask_shape = (
+                coarse_token_ids.shape[0],
+                coarse_token_ids.shape[-1] + fine_token_ids.shape[-1] + 2
+            )
+
+            self_attn_mask = generate_mask_with_prob(mask_shape, self.mask_prob, device = self.device)
 
         coarse_logits, fine_logits = self.transformer(
             coarse_token_ids = coarse_token_ids,
@@ -1753,7 +1877,7 @@ class AudioLM(nn.Module):
 
         self.coarse = CoarseTransformerWrapper(
             wav2vec = wav2vec,
-            codec= codec,
+            codec = codec,
             transformer = coarse_transformer,
             audio_conditioner = audio_conditioner,
             unique_consecutive = unique_consecutive
@@ -1770,14 +1894,16 @@ class AudioLM(nn.Module):
         return next(self.parameters()).device
 
     @eval_decorator
-    @torch.no_grad()
+    @torch.inference_mode()
     def forward(
         self,
         *,
         batch_size = 1,
         text: Optional[List[str]] = None,
-        text_embeds: Optional[torch.Tensor] = None,
+        text_embeds: Optional[Tensor] = None,
         prime_wave = None,
+        prime_wave_input_sample_hz = None,
+        prime_wave_path = None,
         max_length = 2048,
         return_coarse_generated_wave = False,
         mask_out_generated_fine_tokens = False
@@ -1788,19 +1914,31 @@ class AudioLM(nn.Module):
             if exists(text):
                 text_embeds = self.semantic.embed_text(text)
 
+        assert not (exists(prime_wave) and exists(prime_wave_path)), 'prompt audio must be given as either `prime_wave: Tensor` or `prime_wave_path: str`'
+
         if exists(prime_wave):
+            assert exists(prime_wave_input_sample_hz), 'the input sample frequency for the prompt audio must be given as `prime_wave_input_sample_hz: int`'
+            prime_wave = prime_wave.to(self.device)
+        elif exists(prime_wave_path):
+            prime_wave_path = Path(prime_wave_path)
+            assert exists(prime_wave_path), f'file does not exist at {str(prime_wave_path)}'
+
+            prime_wave, prime_wave_input_sample_hz = torchaudio.load(str(prime_wave_path))
             prime_wave = prime_wave.to(self.device)
 
         semantic_token_ids = self.semantic.generate(
             text_embeds = text_embeds if self.semantic_has_condition else None,
             batch_size = batch_size,
             prime_wave = prime_wave,
+            prime_wave_input_sample_hz = prime_wave_input_sample_hz,
             max_length = max_length
         )
 
         coarse_token_ids_or_recon_wave = self.coarse.generate(
             text_embeds = text_embeds if self.coarse_has_condition else None,
             semantic_token_ids = semantic_token_ids,
+            prime_wave = prime_wave,
+            prime_wave_input_sample_hz = prime_wave_input_sample_hz,
             reconstruct_wave = return_coarse_generated_wave
         )
 
@@ -1810,6 +1948,8 @@ class AudioLM(nn.Module):
         generated_wave = self.fine.generate(
             text_embeds = text_embeds if self.fine_has_condition else None,
             coarse_token_ids = coarse_token_ids_or_recon_wave,
+            prime_wave = prime_wave,
+            prime_wave_input_sample_hz = prime_wave_input_sample_hz,
             reconstruct_wave = True,
             mask_out_generated_fine_tokens = mask_out_generated_fine_tokens
         )
