@@ -4,11 +4,16 @@ from einops import rearrange, pack, unpack
 import torch
 from torch import nn
 
+from torchaudio.functional import resample
+
 from vector_quantize_pytorch import ResidualVQ
 
 from encodec import EncodecModel
 from encodec.utils import _linear_overlap_add
 
+def exists(val):
+    return val is not None
+
 class EncodecWrapper(nn.Module):
     """
     Support pretrained 24kHz Encodec by Meta AI, if you want to skip training SoundStream.
@@ -67,15 +72,23 @@ class EncodecWrapper(nn.Module):
     def seq_len_multiple_of(self):
         return reduce(lambda x, y: x * y, self.strides)
 
+    @property
+    def downsample_factor(self):
+        return self.seq_len_multiple_of
+
     def forward(
         self,
         x,
+        input_sample_hz = None,
         return_encoded = False,
         **kwargs
     ):
 
         x, ps = pack([x], '* n')
 
+        if exists(input_sample_hz):
+            x = resample(x, input_sample_hz, self.target_sample_hz)
+
         # kwargs for stuff like return_encoded=True, which SoundStream uses but Encodec doesn't
         assert not self.model.training, "Encodec is pretrained and should never be called outside eval mode."
         # Unlike in the Encodec sample code in its README, x has already been resampled so we don't need to call
@@ -85,12 +98,12 @@ class EncodecWrapper(nn.Module):
         wav = rearrange(x, f'b t -> b {self.model.channels} t')
 
         # Extract discrete codes from EnCodec
-        with torch.no_grad():
+        with torch.inference_mode():
             encoded_frames = self.model.encode(wav)
         # encoded_frames is a list of (frame, scale) tuples. Scale is a scalar but we don't use it. Frame is a tensor
         # of shape [batch, num_quantizers, num_samples_per_frame]. We want to concatenate the frames to get all the
         # timesteps concatenated.
-        codes = torch.cat([encoded[0] for encoded in encoded_frames], dim=1)  # [batch, num_quantizers, timesteps]
+        codes = torch.cat([encoded[0] for encoded in encoded_frames], dim=-1)  # [batch, num_quantizers, timesteps]
         # transformer code that uses codec expects codes to be [batch, timesteps, num_quantizers]
         codes = rearrange(codes, 'b q n -> b n q')  # result: [batch, timesteps, num_quantizers]
         # in original soundstream, is x, indices, commit_loss. But we only use indices in eval mode, so just keep that.
@@ -98,10 +111,11 @@ class EncodecWrapper(nn.Module):
         # allow for returning of sum of quantized embeddings
 
         emb = None
+
         if return_encoded:
             emb = self.get_emb_from_indices(codes)
+            emb, = unpack(emb, ps, '* n c')
 
-        emb, = unpack(emb, ps, '* n c')
         codes, = unpack(codes, ps, '* n q')
 
         return emb, codes, None
