@@ -1,8 +1,8 @@
 from pathlib import Path
 
 import torch
-from torch import nn
-from einops import rearrange, pack, unpack
+from torch import nn, einsum
+from einops import rearrange, repeat, pack, unpack
 
 import joblib
 
@@ -36,6 +36,10 @@ class HubertWithKmeans(nn.Module):
         output_layer = 9
     ):
         super().__init__()
+
+        import sklearn
+        assert sklearn.__version__ == '0.24.0', 'scikit-learn needs to be exactly 0.24.0 - please install the correct version by running `pip install scikit-learn==0.24.0`'
+
         self.target_sample_hz = target_sample_hz
         self.seq_len_multiple_of = seq_len_multiple_of
         self.output_layer = output_layer
@@ -54,8 +58,14 @@ class HubertWithKmeans(nn.Module):
         self.model.eval()
 
         kmeans = joblib.load(kmeans_path)
+
         self.kmeans = kmeans
 
+        self.register_buffer(
+            'cluster_centers',
+            torch.from_numpy(kmeans.cluster_centers_)
+        )
+
     @property
     def groups(self):
         return 1
@@ -64,14 +74,19 @@ class HubertWithKmeans(nn.Module):
     def codebook_size(self):
         return self.kmeans.n_clusters
 
-    @torch.no_grad()
+    @property
+    def downsample_factor(self):
+        # todo: double check
+        return 320
+
+    @torch.inference_mode()
     def forward(
         self,
         wav_input,
         flatten = True,
         input_sample_hz = None
     ):
-        device = wav_input.device
+        batch, device = wav_input.shape[0], wav_input.device
 
         if exists(input_sample_hz):
             wav_input = resample(wav_input, input_sample_hz, self.target_sample_hz)
@@ -84,16 +99,13 @@ class HubertWithKmeans(nn.Module):
             features_only = True,
             mask = False,  # thanks to @maitycyrus for noticing that mask is defaulted to True in the fairseq code
             output_layer = self.output_layer
-        )
+        )['x']
 
-        embed, packed_shape = pack([embed['x']], '* d')
-
-        codebook_indices = self.kmeans.predict(embed.cpu().detach().numpy())
-
-        codebook_indices = torch.from_numpy(codebook_indices).to(device).long()
+        batched_cluster_centers = repeat(self.cluster_centers, 'c d -> b c d', b = embed.shape[0])
+        dists = -torch.cdist(embed, batched_cluster_centers, p = 2)
+        clusters = dists.argmax(dim = -1)
 
         if flatten:
-            return codebook_indices
+            return clusters
 
-        codebook_indices, = unpack(codebook_indices, packed_shape, '*')
-        return codebook_indices
+        return rearrange(clusters, 'b ... -> b (...)')
