@@ -88,35 +88,38 @@ class MultiScaleDiscriminator(nn.Module):
         self,
         channels = 16,
         layers = 4,
-        groups = 4,
+        groups = (4, 16, 64, 256),
         chan_max = 1024,
         input_channels = 1
     ):
         super().__init__()
-        self.init_conv = nn.Conv1d(input_channels, channels, 7)
+        self.init_conv = nn.Conv1d(input_channels, channels, 15, padding = 7)
         self.conv_layers = nn.ModuleList([])
 
         curr_channels = channels
 
-        for _ in range(layers):
+        for _, group in zip(range(layers), groups):
             chan_out = min(curr_channels * 4, chan_max)
 
             self.conv_layers.append(nn.Sequential(
-                nn.Conv1d(curr_channels, chan_out, 8, stride = 4, padding = 4, groups = groups),
+                nn.Conv1d(curr_channels, chan_out, 41, stride = 4, padding = 20, groups = group),
                 leaky_relu()
             ))
 
             curr_channels = chan_out
 
         self.final_conv = nn.Sequential(
-            nn.Conv1d(curr_channels, curr_channels, 3),
+            nn.Conv1d(curr_channels, curr_channels, 5, padding = 2),
             leaky_relu(),
-            nn.Conv1d(curr_channels, 1, 1),
+            nn.Conv1d(curr_channels, 1, 3, padding = 1),
         )
 
-    def forward(self, x, return_intermediates = False):
+    def forward(
+        self,
+        x,
+        return_intermediates = False
+    ):
         x = self.init_conv(x)
-
         intermediates = []
 
         for layer in self.conv_layers:
@@ -428,6 +431,8 @@ class SoundStream(nn.Module):
         rq_ema_decay = 0.95,
         rq_quantize_dropout_multiple_of = 1,
         rq_groups = 1,
+        rq_stochastic_sample_codes = False,
+        rq_kwargs: dict = {},
         input_channels = 1,
         discr_multi_scales = (1, 0.5, 0.25),
         stft_normalized = False,
@@ -519,7 +524,9 @@ class SoundStream(nn.Module):
             kmeans_init = True,
             threshold_ema_dead_code = 2,
             quantize_dropout = True,
-            quantize_dropout_cutoff_index = quantize_dropout_cutoff_index
+            quantize_dropout_cutoff_index = quantize_dropout_cutoff_index,
+            stochastic_sample_codes = rq_stochastic_sample_codes,
+            **rq_kwargs
         )
 
         self.decoder_film = FiLM(codebook_dim, dim_cond = 2)
@@ -676,6 +683,10 @@ class SoundStream(nn.Module):
     def seq_len_multiple_of(self):
         return functools.reduce(lambda x, y: x * y, self.strides)
 
+    @property
+    def downsample_factor(self):
+        return self.seq_len_multiple_of
+
     def process_input(
         self,
         x,
