@@ -43,13 +43,22 @@ from audiolm_pytorch.utils import AudioConditionerBase
 from audiolm_pytorch.version import __version__
 from packaging import version
 
-from accelerate import Accelerator
+from accelerate import (Accelerator, DistributedType)
 from accelerate.utils import DistributedDataParallelKwargs
 
 # constants
 
 DEFAULT_SAMPLE_RATE = 16000
 
+# make sure only one trainer is instantiated
+
+ONE_TRAINER_INSTANTIATED = False
+
+def check_one_trainer():
+    global ONE_TRAINER_INSTANTIATED
+    assert not ONE_TRAINER_INSTANTIATED, 'only one Trainer can be instantiated at a time for training'
+    ONE_TRAINER_INSTANTIATED = True
+
 # for automatically routing data emitted from a dataset to keywords of the transformer wrappers
 
 DATASET_FIELD_TYPE_CONFIG = dict(
@@ -118,7 +127,12 @@ def checkpoint_num_steps(checkpoint_path):
     Filename format assumed to be something like "/path/to/semantic.transformer.20000.pt" which is
     for 20k train steps. Returns 20000 in that case.
     """
-    return int(re.findall(r'\d+', str(checkpoint_path))[-1])
+    results = re.findall(r'\d+', str(checkpoint_path))
+
+    if len(results) == 0:
+        return 0
+
+    return int(results[-1])
 
 # main trainer class
 
@@ -154,6 +168,8 @@ class SoundStreamTrainer(nn.Module):
         dl_num_workers: int = 0,
         accelerator: Accelerator = None,
         accelerate_kwargs: dict = dict(),
+        dataloader_drop_last = True,
+        split_batches = False,
         use_lion: bool = False,
         force_clear_prev_results: bool = None  # set to True | False to skip the prompt
     ):
@@ -162,13 +178,19 @@ class SoundStreamTrainer(nn.Module):
         train/val DataLoader instances.
         """
         super().__init__()
+        check_one_trainer()
 
         if accelerator:
             self.accelerator = accelerator
             assert len(accelerate_kwargs) == 0
         else:
             kwargs = DistributedDataParallelKwargs(find_unused_parameters = True)
-            self.accelerator = Accelerator(kwargs_handlers = [kwargs], **accelerate_kwargs)
+
+            self.accelerator = Accelerator(
+                kwargs_handlers = [kwargs],
+                split_batches = split_batches,
+                **accelerate_kwargs
+            )
 
         self.soundstream = soundstream
 
@@ -242,11 +264,14 @@ class SoundStreamTrainer(nn.Module):
                 self.valid_ds = self.ds
                 self.print(f'training with shared training and valid dataset of {len(self.ds)} samples')
 
+            assert len(self.ds) >= batch_size, 'dataset must have sufficient samples for training'
+            assert len(self.valid_ds) >= batch_size, f'validation dataset must have sufficient number of samples (currently {len(self.valid_ds)}) for training'
+
             # dataloader
 
-            self.dl = get_dataloader(self.ds, batch_size = batch_size, num_workers = dl_num_workers, shuffle = True)
+            self.dl = get_dataloader(self.ds, batch_size = batch_size, num_workers = dl_num_workers, shuffle = True, drop_last = dataloader_drop_last)
 
-            self.valid_dl = get_dataloader(self.valid_ds, batch_size = batch_size, num_workers = dl_num_workers, shuffle = True)
+            self.valid_dl = get_dataloader(self.valid_ds, batch_size = batch_size, num_workers = dl_num_workers, shuffle = True, drop_last = dataloader_drop_last)
 
         # prepare with accelerator
 
@@ -288,9 +313,12 @@ class SoundStreamTrainer(nn.Module):
         self.results_folder.mkdir(parents = True, exist_ok = True)
 
         # Initialize experiment trackers if an external Accelerator is not passed in
+
         if not accelerator:
             self.accelerator.init_trackers("soundstream", config=hyperparameters)        
 
+        assert self.accelerator.distributed_type != DistributedType.FSDP, 'FSDP not supported for soundstream trainer due to complex-valued stft discriminator'
+
     def set_model_as_ema_model_(self):
         """ this will force the main 'online' model to have same parameters as the exponentially moving averaged model """
         assert self.use_ema
@@ -351,7 +379,9 @@ class SoundStreamTrainer(nn.Module):
         for key, _ in self.multiscale_discriminator_iter():
             discr_optim = getattr(self, key)
             discr_optim.load_state_dict(pkg[key])
+
         # + 1 to start from the next step and avoid overwriting the last checkpoint
+
         self.steps = torch.tensor([checkpoint_num_steps(path) + 1], device=self.device)
 
     def multiscale_discriminator_iter(self):
@@ -504,7 +534,7 @@ class SoundStreamTrainer(nn.Module):
             for model, label in models:
                 model.eval()
 
-                with torch.no_grad():
+                with torch.inference_mode():
                     recons = model(wave, return_recons_only = True)
 
                 for ind, recon in enumerate(recons.unbind(dim = 0)):
@@ -560,10 +590,18 @@ class SemanticTransformerTrainer(nn.Module):
         save_model_every = 1000,
         results_folder = './results',
         accelerate_kwargs: dict = dict(),
-        force_clear_prev_results = None
+        split_batches = False,
+        drop_last = False,
+        force_clear_prev_results = None,
+        average_valid_loss_over_grad_accum_every: bool = True, # if False, valid loss on a single batch
     ):
         super().__init__()
-        self.accelerator = Accelerator(**accelerate_kwargs)
+        check_one_trainer()
+
+        self.accelerator = Accelerator(
+            split_batches = split_batches,
+            **accelerate_kwargs
+        )
 
         self.wav2vec = wav2vec
         self.transformer = transformer
@@ -620,11 +658,14 @@ class SemanticTransformerTrainer(nn.Module):
             self.valid_ds = self.ds
             self.print(f'training with shared training and valid dataset of {len(self.ds)} samples')
 
+        assert len(self.ds) >= batch_size, 'dataset must have sufficient samples for training'
+        assert len(self.valid_ds) >= batch_size, f'validation dataset must have sufficient number of samples (currently {len(self.valid_ds)}) for training'
+
         # dataloader
 
-        self.dl = get_dataloader(self.ds, batch_size = batch_size, shuffle = True)
+        self.dl = get_dataloader(self.ds, batch_size = batch_size, shuffle = True, drop_last = drop_last)
 
-        self.valid_dl = get_dataloader(self.valid_ds, batch_size = batch_size, shuffle = True)
+        self.valid_dl = get_dataloader(self.valid_ds, batch_size = batch_size, shuffle = True, drop_last = drop_last)
 
         # prepare with accelerator
 
@@ -657,6 +698,7 @@ class SemanticTransformerTrainer(nn.Module):
         
         hps = {"num_train_steps": num_train_steps, "data_max_length": data_max_length, "learning_rate": lr}
         self.accelerator.init_trackers("semantic", config=hps)
+        self.average_valid_loss_over_grad_accum_every = average_valid_loss_over_grad_accum_every
 
     def save(self, path):
         pkg = dict(
@@ -667,18 +709,11 @@ class SemanticTransformerTrainer(nn.Module):
         torch.save(pkg, path)
 
     def load(self, path):
-        path = Path(path)
-        assert path.exists()
-        pkg = torch.load(str(path), map_location = 'cpu')
-
-        # check version
-
-        if 'version' in pkg and version.parse(pkg['version']) < version.parse(__version__):
-            print(f'model was trained on older version {pkg["version"]} of audiolm-pytorch')
-
         transformer = self.accelerator.unwrap_model(self.transformer)
-        transformer.load_state_dict(pkg['model'])
+        pkg = transformer.load(path)
+        # trainer-specific things
         self.optim.load_state_dict(pkg['optim'])
+
         # + 1 to start from the next step and avoid overwriting the last checkpoint
         self.steps = torch.tensor([checkpoint_num_steps(path) + 1], device=self.device)
 
@@ -747,13 +782,18 @@ class SemanticTransformerTrainer(nn.Module):
 
         # sample results every so often
 
+        self.accelerator.wait_for_everyone()
+
         if self.is_main and not (steps % self.save_results_every):
-            data_kwargs = self.data_tuple_to_kwargs(next(self.valid_dl_iter))
-
-            with torch.no_grad():
-                self.train_wrapper.eval()
-                valid_loss = self.train_wrapper(**data_kwargs, return_loss = True)
+            valid_loss = 0
+            for _ in range(self.average_valid_loss_over_grad_accum_every):
+                data_kwargs = self.data_tuple_to_kwargs(next(self.valid_dl_iter))
 
+                with torch.inference_mode():
+                    self.train_wrapper.eval()
+                    valid_loss += self.train_wrapper(**data_kwargs, return_loss = True)
+            valid_loss = valid_loss.clone() # avoid inference mode to non-inference mode error
+            valid_loss /= self.average_valid_loss_over_grad_accum_every
             self.print(f'{steps}: valid loss {valid_loss}')
             self.accelerator.log({"valid_loss": valid_loss}, step=steps)
 
@@ -804,10 +844,18 @@ class CoarseTransformerTrainer(nn.Module):
         save_model_every = 1000,
         results_folder = './results',
         accelerate_kwargs: dict = dict(),
-        force_clear_prev_results = None
+        split_batches = False,
+        drop_last = False,
+        force_clear_prev_results = None,
+        average_valid_loss_over_grad_accum_every: bool = True,  # if False, valid loss on a single batch
     ):
         super().__init__()
-        self.accelerator = Accelerator(**accelerate_kwargs)
+        check_one_trainer()
+
+        self.accelerator = Accelerator(
+            split_batches = split_batches,
+            **accelerate_kwargs
+        )
 
         self.transformer = transformer
         self.codec = codec
@@ -870,11 +918,14 @@ class CoarseTransformerTrainer(nn.Module):
             self.valid_ds = self.ds
             self.print(f'training with shared training and valid dataset of {len(self.ds)} samples')
 
+        assert len(self.ds) >= batch_size, 'dataset must have sufficient samples for training'
+        assert len(self.valid_ds) >= batch_size, f'validation dataset must have sufficient number of samples (currently {len(self.valid_ds)}) for training'
+
         # dataloader
 
-        self.dl = get_dataloader(self.ds, batch_size = batch_size, shuffle = True)
+        self.dl = get_dataloader(self.ds, batch_size = batch_size, shuffle = True, drop_last = drop_last)
 
-        self.valid_dl = get_dataloader(self.valid_ds, batch_size = batch_size, shuffle = True)
+        self.valid_dl = get_dataloader(self.valid_ds, batch_size = batch_size, shuffle = True, drop_last = drop_last)
 
         # prepare with accelerator
 
@@ -909,6 +960,7 @@ class CoarseTransformerTrainer(nn.Module):
         self.accelerator.init_trackers("coarse", config=hps)        
 
         self.train_wrapper.to(self.device)
+        self.average_valid_loss_over_grad_accum_every = average_valid_loss_over_grad_accum_every
 
     def save(self, path):
         pkg = dict(
@@ -919,19 +971,11 @@ class CoarseTransformerTrainer(nn.Module):
         torch.save(pkg, path)
 
     def load(self, path):
-        path = Path(path)
-        assert path.exists()
-        pkg = torch.load(str(path), map_location = 'cpu')
-
-        # check version
-
-        if 'version' in pkg and version.parse(pkg['version']) < version.parse(__version__):
-            print(f'model was trained on older version {pkg["version"]} of audiolm-pytorch')
-
         transformer = self.accelerator.unwrap_model(self.transformer)
-        transformer.load_state_dict(pkg['model'])
-
+        pkg = transformer.load(path)
+        # trainer-specific things
         self.optim.load_state_dict(pkg['optim'])
+
         # + 1 to start from the next step and avoid overwriting the last checkpoint
         self.steps = torch.tensor([checkpoint_num_steps(path) + 1], device=self.device)
 
@@ -996,17 +1040,22 @@ class CoarseTransformerTrainer(nn.Module):
 
         # sample results every so often
 
+        self.accelerator.wait_for_everyone()
+
         if self.is_main and not (steps % self.save_results_every):
-            data_kwargs = dict(zip(self.ds_fields, next(self.valid_dl_iter)))
+            valid_loss = 0
+            for i in range(self.average_valid_loss_over_grad_accum_every):
+                data_kwargs = dict(zip(self.ds_fields, next(self.valid_dl_iter)))
 
-            with torch.no_grad():
-                self.train_wrapper.eval()
-
-                valid_loss = self.train_wrapper(
-                    **data_kwargs,
-                    return_loss = True
-                )
+                with torch.inference_mode():
+                    self.train_wrapper.eval()
 
+                    valid_loss += self.train_wrapper(
+                        **data_kwargs,
+                        return_loss = True
+                    )
+            valid_loss = valid_loss.clone() # avoid inference mode to non-inference mode error
+            valid_loss /= self.average_valid_loss_over_grad_accum_every
             self.print(f'{steps}: valid loss {valid_loss}')
             self.accelerator.log({"valid_loss": valid_loss}, step=steps)
 
@@ -1056,10 +1105,18 @@ class FineTransformerTrainer(nn.Module):
         save_model_every = 1000,
         results_folder = './results',
         accelerate_kwargs: dict = dict(),
-        force_clear_prev_results = None
+        split_batches = False,
+        drop_last = False,
+        force_clear_prev_results = None,
+        average_valid_loss_over_grad_accum_every: bool = True, # if False, valid loss on a single batch
     ):
         super().__init__()
-        self.accelerator = Accelerator(**accelerate_kwargs)
+        check_one_trainer()
+
+        self.accelerator = Accelerator(
+            split_batches = split_batches,
+            **accelerate_kwargs
+        )
 
         self.transformer = transformer
         self.codec = codec
@@ -1117,11 +1174,14 @@ class FineTransformerTrainer(nn.Module):
             self.valid_ds = self.ds
             self.print(f'training with shared training and valid dataset of {len(self.ds)} samples')
 
+        assert len(self.ds) >= batch_size, 'dataset must have sufficient samples for training'
+        assert len(self.valid_ds) >= batch_size, f'validation dataset must have sufficient number of samples (currently {len(self.valid_ds)}) for training'
+
         # dataloader
 
-        self.dl = get_dataloader(self.ds, batch_size = batch_size, shuffle = True)
+        self.dl = get_dataloader(self.ds, batch_size = batch_size, shuffle = True, drop_last = drop_last)
 
-        self.valid_dl = get_dataloader(self.valid_ds, batch_size = batch_size, shuffle = True)
+        self.valid_dl = get_dataloader(self.valid_ds, batch_size = batch_size, shuffle = True, drop_last = drop_last)
 
         # prepare with accelerator
 
@@ -1156,6 +1216,7 @@ class FineTransformerTrainer(nn.Module):
         self.accelerator.init_trackers("fine", config=hps)        
 
         self.train_wrapper.to(self.device)
+        self.average_valid_loss_over_grad_accum_every = average_valid_loss_over_grad_accum_every
 
     def save(self, path):
         pkg = dict(
@@ -1166,23 +1227,14 @@ class FineTransformerTrainer(nn.Module):
         torch.save(pkg, path)
 
     def load(self, path):
-        path = Path(path)
-        assert path.exists()
-        pkg = torch.load(str(path), map_location = 'cpu')
-
-        # check version
-
-        if 'version' in pkg and version.parse(pkg['version']) < version.parse(__version__):
-            print(f'model was trained on older version {pkg["version"]} of audiolm-pytorch')
-
         transformer = self.accelerator.unwrap_model(self.transformer)
-        transformer.load_state_dict(pkg['model'])
-
+        pkg = transformer.load(path)
+        # trainer-specific things
         self.optim.load_state_dict(pkg['optim'])
+
         # + 1 to start from the next step and avoid overwriting the last checkpoint
         self.steps = torch.tensor([checkpoint_num_steps(path) + 1], device=self.device)
 
-
     def print(self, msg):
         self.accelerator.print(msg)
 
@@ -1246,13 +1298,18 @@ class FineTransformerTrainer(nn.Module):
 
         # sample results every so often
 
+        self.accelerator.wait_for_everyone()
+
         if self.is_main and not (steps % self.save_results_every):
-            data_kwargs = self.data_tuple_to_kwargs(next(self.valid_dl_iter))
-
-            with torch.no_grad():
-                self.train_wrapper.eval()
-                valid_loss = self.train_wrapper(**data_kwargs, return_loss = True)
+            valid_loss = 0
+            for i in range(self.average_valid_loss_over_grad_accum_every):
+                data_kwargs = self.data_tuple_to_kwargs(next(self.valid_dl_iter))
 
+                with torch.inference_mode():
+                    self.train_wrapper.eval()
+                    valid_loss += self.train_wrapper(**data_kwargs, return_loss = True)
+            valid_loss = valid_loss.clone() # avoid inference mode to non-inference mode error
+            valid_loss /= self.average_valid_loss_over_grad_accum_every
             self.print(f'{steps}: valid loss {valid_loss}')
             self.accelerator.log({"valid_loss": valid_loss}, step=steps)
 
@@ -1273,4 +1330,4 @@ class FineTransformerTrainer(nn.Module):
             logs = self.train_step()
             log_fn(logs)
 
-        self.print('training complete')
+        self.print('training complete')
\ No newline at end of file
